"""Module of transition tensors"""

import jax.numpy as jnp
from abc import ABC, abstractmethod

# TODO: consider implementation with Tucker decomposition or other tensor decompositions (SVD/PCA)
#       ... to implement the independent and dependent transition kernels separately
#       ... similar concepts from Dynamic Bayesian Networks, exponential families

class transition_tensor(ABC):
    def __init__(self, num_latent:int)->None:
        self.num_latent = num_latent

    @abstractmethod
    def t_tensor(self)->jnp.ndarray:
        pass
    
class poisson_arrival(transition_tensor):
    """
    Transition happens with Poisson arrivals. State value is drawn by a marginal probability on states.
    The poisson arrivals are geometrically spaced.
    From Markov Switching Multifractal model. 
    """

    def __init__(self,
                num_latent:int,
                marg_prob_mass:jnp.ndarray
                ) -> None:
        super().__init__(num_latent)
        self.marg_prob_mass = marg_prob_mass

    #NOTE: currently the forward algorithm breaks if given a jax tensor as input, due to the way enumerate handles the predictive update
    def t_tensor(self, arrival_gdistance:float, hf_arrival:float)->tuple:
        arrivals = 1 - (1 - hf_arrival) ** (1 / (arrival_gdistance ** (jnp.arange(self.num_latent, 0, -1) - 1)))
        len_pm = len(self.marg_prob_mass)
        return tuple((1-g)*jnp.eye(len_pm) + g*jnp.tile(self.marg_prob_mass, (len_pm, 1)) for g in arrivals)


# Mid-level classes for Hidden Markov Models

# TODO: 
# proper math rendering
# GENERALIZE discrete_distribution to any discrete with dependence adjustment factor
# handling pandas inputs (e.g. py_trees)

# NOTE:
# __init__ of the components sets the hyperparameters, the specific functions take parameters as input (not stored as self instances)
# the abstract HMM calls the abstract methods in __init__, 
# ... the concretization should input the concrete version, along with the proper initialization inputs

import jax.numpy as jnp
import numpy as np

from abc import abstractmethod #temporary

from fractrics._ts_components._core import stochastic_ts

import fractrics._ts_components._HMM._forward as _forward
import fractrics._ts_components._HMM._data_likelihood as _data_likelihood
import fractrics._ts_components._HMM._initial_distribution as _initial_distribution
import fractrics._ts_components._HMM._transition_tensor as _transition_tensor

class HMM(stochastic_ts):
    """Generic class for Hidden Markov Models."""

    def __init__(self, ts: np.ndarray | jnp.ndarray,
                 forward: _forward.forward_update,
                 data_likelihood: _data_likelihood.data_likelihood,
                 initial_dist: _initial_distribution.homogeneous_independent,
                 transition_tensor: _transition_tensor.transition_tensor,
                 name: str | None = None,
                 num_latent: int = 1) -> None:
        
        super().__init__(ts, name)

        self.num_latent = num_latent
        self._initial_dist = initial_dist
        self._data_likelihood = data_likelihood
        self._transition_tensor = transition_tensor
        self._forward = forward
    
    def _MPM_states(self, states: jnp.ndarray, distr_list:jnp.ndarray)-> jnp.ndarray:
        """
        At each t returns the states corresponding to the maximized posterior disjoint marginal.
        distr_list: matrix where each column is the distribution of latent states at a given time, result of forward algorithm.
        """
        return states[jnp.argmax(distr_list, axis=1)]
    
    def _backward(self)-> jnp.ndarray:
        raise NotImplementedError
    
    def _viterbi(self)-> jnp.ndarray:
        """Computes the most probable path of latent states (joint probability)"""
        raise NotImplementedError

    # TODO: currently defined for the factor HMM, restructure for working in general (most of the code is the same)
    @abstractmethod
    def forecast(self, horizon):
        super().forecast(filtration=horizon)



from fractrics._deprecated._pending_refactor._helper import OLS, partition_map
import numpy as np
import pandas as pd
import scipy.special as sp
from scipy.stats import norm, chi2, t, f

def z_score(array):
    """ 
    Computes the Z-score for an input array
    """
    
    # ensure to work with a numpy array
    array = np.array(array)
    return (array - np.nanmean(array))/np.nanstd(array)

def kurtosis(array):
    """
    Computes kurtosis of input array
    """
    return (z_score(array)**4).mean()

def skewness(array):
    """
    Computes skewness of input array
    """

    return (z_score(array)**3).mean()

def acf(ts):
    """
    Compute autocorrelation, T-ratio and Ljung-Box and their p-values for all lags of a time series.
    Parameters:
        ts: Time series data.

    Returns:
        acf (ndarray): Autocorrelation values and statistics for lags 0 to n-1.
    """
    
    ts = np.asarray(ts)
    n = len(ts)
    lags = range(n)
    acf = (np.correlate(ts - np.nanmean(ts), ts - np.nanmean(ts), mode='full'))[n-1:] / (np.nanvar(ts)*n)
    t_ratio = acf/(np.sqrt(1+2*np.cumsum(acf**2)))
    pv_t = 2 * (1 -norm.cdf(abs(t_ratio)))
    q_st = n*(n+2)*np.cumsum(acf**2/np.arange(n, 0, -1))
    pv_q = 1 - chi2.cdf(q_st, lags)
        
    acf_df = pd.DataFrame({
    'lag': lags,
    'acf': acf,
    't_ratio': t_ratio,
    'pv_t' : pv_t,
    'q_stat': q_st,
    'pv_q': pv_q
    })
    
    return acf_df

def rs(ts, logc=0, even=True):
    """
    Computes the rescaled range statistic for a time series.
    
    :param ts: the series to be analized
    :param even: whether to compute the rescaled range only on evenly split partitions
    :param logc: specifies the order of the log change of the series to be used. (0: no log change, 1: first order log change)
    """
    ts = np.array(ts) # making sure the series is a numpy array
    
    ts_adj = np.array(np.log(ts[1:] / (ts[:-1]**logc))) #transform in logaritmic scale
    
    # all lengths of partitions up to 2 that divide the series in equal parts
    if even: n = np.array([n for n in range(2, ((len(ts)-1)//2)+1) if (len(ts)-1) % n == 0])
    else: n = np.arange(2, ((len(ts)-1)//2)+1)
    
    rs = np.zeros(len(n)) #initialize vector of R/S
    r = np.zeros(len(n)) #initialize vector of adjusted range
    
    # expected statistics of a brownian motion series according to Anis and Lloyd
    i = np.arange(1, n[-1])
    e_rs = (1/np.pi)*(np.exp(sp.gammaln(0.5*(n-1))-sp.gammaln(0.5*n)))*np.array([np.sum(((k - i[:k-1]) / i[:k-1]) ** 0.5) for k in n])

    def rs_stat(array):
        
        # cumulative deviance from mean
        cum_dev = np.cumsum(array - np.nanmean(array))

        return (np.nanmax(cum_dev)-np.nanmin(cum_dev))/ np.nanstd(array)
    
    def r_stat(array):
    
        # cumulative deviance from mean
        cum_dev = np.cumsum(array - np.nanmean(array))

        return (np.nanmax(cum_dev)-np.nanmin(cum_dev))
    
    for j,i in enumerate(n):
        # each element is the average R/S statistic for the corresponding number of days
        rs[j] = np.nanmean(partition_map(arr=ts_adj, func=rs_stat, num_partitions=i))
        r[j] = np.nanmean(partition_map(arr=ts_adj, func=r_stat, num_partitions=i))
        
    # note: (number of partitions) = (size of the ts) / (days in each partitions)
    # since n contains all possible integer numbers of partitions, it is the inverse of an array containing the number of days in each partitions
    # therefore n is inverted to be used in a log plot. The same is dove for e_rs
    
    rs_df = pd.DataFrame({
        'n_days': n[::-1], # reverse the array to resemble the number of days
        'log_days': np.log(n[::-1]),
        
        'rs': rs,
        'e_rs': e_rs[::-1],
        'loge_rs': np.log(e_rs[::-1]),
        'log_rs': np.log(rs),
        'v_stat': rs / np.sqrt(n[::-1]),
        
        'r':r,
        'e_r': np.sqrt(n[::-1]*np.pi/2),
        'var_r': np.sqrt(n[::-1]*(np.pi**2/6 - np.pi/2)),
        'z_r': (r - np.sqrt(n[::-1]*np.pi/2))/np.sqrt(n[::-1]*(np.pi**2/6 - np.pi/2)),
        'pv_r': 2*(1- norm.cdf(abs((r - np.sqrt(n[::-1]*np.pi/2))/np.sqrt(n[::-1]*(np.pi**2/6 - np.pi/2)))))
        })
    
    return rs_df

def hurst(ts, logc=0, even=True):
    """
    Returns the hurst exponent of a series, its gaussian z-score and p-value,
    and the minimum number of observations to consider the result significant at the 5% 
    confidence interval
    """
    ts_rs = rs(ts=ts, logc=logc, even=even)
    h = OLS(x=ts_rs['log_days'], y=ts_rs['log_rs'], intercept=True)
    e_h = OLS(x=ts_rs['log_days'], y=ts_rs['loge_rs'], intercept=True)
   
    # minimum observations to consider an H != 0.5 significant
    min_T = 4/((h[0]-e_h[0])**2)
    gauss_z = (h[0]-e_h[0])*np.sqrt(len(ts))
    pv_h = 2 *(1 - norm.cdf(gauss_z))
    
    #OLS predictions and residuals
    h_fit = ts_rs['log_days']*h[0]+h[1]
    h_resid = ts_rs['log_rs'] - h_fit
    
    #variance of the residuals of an OLS: -2 parameters H and c
    deg_freedom = (len(h_resid)-2)
    resid_var = np.sum(h_resid**2)/deg_freedom
    
    #standard error of each parameter
    std_err = np.sqrt(resid_var/h)
    
    # T statistics and p. value
    t_stat = h/std_err
    pv_t = 2 * (1 - t.cdf(np.abs(t_stat), deg_freedom))
    
    # F statistics   
    f_stat = np.sum((h_fit - ts_rs['log_rs'].mean())**2)/resid_var
    pv_f = f.cdf(f_stat, 1, deg_freedom)
    
    hurst_={
        "hurst": h[0],
        "c": h[1],
        "pv_hurst": pv_h,
        "min_obs": min_T,
        "pv_t":pv_t,
        "pv_F":pv_f
    }
    
    return hurst_




import jax
import itertools
import numpy as np
import jax.numpy as jnp
from jaxopt import LBFGSB
from functools import partial
import jax.scipy.stats as jss
from jaxopt import ScipyMinimize
from scipy.stats import rv_discrete
from jax.scipy.special import logsumexp
from fractrics._deprecated._pending_refactor._diagnostics import pareto_shape, tail_pv

def sim_levelMSM(support, prob, sigma_l, b, gamma_k, k, beta0, beta1, eta, T, init):
    """
    Simulates a Markov Switching Multifractal with level effect
    """
    support = np.array(support)
    prob = np.array(prob)
    
    # construct the elements for the simulation
    M = rv_discrete(values=(support, prob))
    gamma = 1 - (1 - gamma_k) ** (1 / (b ** (np.arange(k, 0, -1) - 1)))
    
    #initialize simulated state and series vectors
    r_sim = np.zeros(T+1)
    x_sim = np.zeros(T)
    v_sim = np.zeros(T)
    state = M.rvs(size=k)
    r_new = init
    #simulation
    for t in range(T):
        r_sim[t] = r_new
        # draw which of the states will transition at time t
        switch = np.random.rand(len(gamma)) < gamma
        
        # using mask switch, draw from M only the new states
        state[switch] = M.rvs(size=np.sum(switch))
        
        # the simulated volatility and return are
        v_sim[t]=sigma_l*np.sqrt(np.prod(state))
        x_sim[t]=v_sim[t]*np.random.standard_normal()
        
        r_new = beta0 + r_sim[t]*(x_sim[t]*r_sim[t]**(eta-1) + 1 + beta1)
        
    return r_sim, v_sim

class level_MSM:
    
    def __init__(self, ts, k=None, pM=None):
        """
        Level-Markov Switching  Multifractal initialization.

        :param ts: time series data (array-like)
        :param k: number of possible states
        :param pM: probability vector of the discrete marginal distribution of multipliers. Default is symmetric Bernoulli.
        
        """
        self.ts = jnp.array(ts)
        self.dts = self.ts[1:]-self.ts[:-1]
        self.k = 2 if k is None else k
        self.pM = jnp.repeat(0.5, 2) if pM is None else jnp.array(pM)
        self.fitted = False

    @partial(jax.jit, static_argnames=["self"])
    def ergotic_distr(self, support):
        """
        Log of Ergodic distribution and possible states (non log).
        States is the k-th repeated cartesian product of support
        The ergotic distribution is vector resulting from row-multiplication of the k-th cartesian product of pM
        """
        states = jnp.array(list(itertools.product(support, repeat=self.k)))
        prob_cp = jnp.array(list(itertools.product(self.pM, repeat=self.k)))
        probs = jnp.prod(prob_cp, axis=1).reshape([len(self.pM)] * self.k)
        return states, jnp.log(probs)
    
    @partial(jax.jit, static_argnames=["self"])
    def w(self, x, states, sigma_l):
        """Likelihood vector for observed data and state permutations."""
        sigma = sigma_l * jnp.sqrt(jnp.prod(states, axis=1))
        return jss.norm.logpdf(x.reshape(-1, 1), loc=0, scale=sigma)

    @partial(jax.jit, static_argnames=["self"])
    def gamma_prob(self, b, gamma_k):
        """Transition probabilities for each k."""
        return 1 - (1 - gamma_k) ** (1 / (b ** (jnp.arange(self.k, 0, -1) - 1)))

    @partial(jax.jit, static_argnames=["self"])
    def A_tensor(self, gamma_vec):
        """
        Computes the transition tensor of the chain. 
        Each element is P[M_(k, t) = j | M_(k, t-1) = i]
        """
        A_ten = jnp.log(jnp.array([(1-g)*jnp.eye(len(self.pM)) + g*jnp.tile(self.pM, (len(self.pM), 1)) for g in gamma_vec]))
        return A_ten

    @partial(jax.jit, static_argnames=["self"])
    def _x(self, beta0, beta1, eta, newd=None, oldd=None):
        """
        Model-implied multifractal component in log-space.
        """
        if newd is None:
            # Sign decomposition trick to keep the log consistent
            den = self.ts[1:]
            num = self.dts - beta0 - beta1 * self.ts[1:]
            sign = jnp.sign(num)*jnp.sign(den)
            
        else:
            # ensure new and old observation are arrays even if only 1 value is passed.
            newd = jnp.atleast_1d(newd)
            oldd = jnp.atleast_1d(oldd)
            
            # if the above have 1 element = oldd, else concatenates last oldd to all but last newd
            den = jnp.concatenate([oldd[-1][None], newd[:-1]])
            new_dts = newd - den          
            num = new_dts - beta0 - beta1 * den
            sign = jnp.sign(num)*jnp.sign(den)
        
        x = sign * jnp.exp(jnp.log(jnp.abs(num)) - eta*jnp.log(jnp.abs(den)))
        return x
    
    @partial(jax.jit, static_argnames=["self", "comp"])
    def update_distr(self, sigma_l, b, gamma_k, support, beta0, beta1, eta, newd=None, oldd=None, comp=True, pi_start=None):
        """
        Recursively compute the state probabilities over time and the NLL of the chain.

        :param gamma_k: transition probability base
        :param b: scaling parameter for transition probabilities
        :param sigma_l: volatility scalar
        :param newd: (used only if fit=True) new data to update the distribution to (for online learning)
        :param comp: specifies whether to return only the likelihood (comp=False)
        :param pi_start: tensor from which to start the update
        :return: final state probability tensor, previous state tensors, Log Likelihood (LL)
        """
         
        def transition_tensor(carry, wr):
            w_t, re_t = wr
            pi_t, LL = carry

            # Apply the transition tensor to get the updated state probabilities
            for k in range(A_tens.shape[0]):
                # Move mode-k axis to front
                log_A_k = A_tens[k].T
                pi_t = jnp.moveaxis(pi_t, k, 0)

                # For each j, sum over i: logsumexp(log_A_k[i, j] + log_pi_t[i])
                pi_t = jax.vmap(  
                    lambda log_col: logsumexp(log_col[:, None] + pi_t, axis=0)
                )(log_A_k)

                # Move the axis back
                pi_t = jnp.moveaxis(pi_t, 0, k)
                
            pi_adj = pi_t.flatten() + w_t
            
            #differently from MSM, the likelihood does not scale the distribution directly
            den = -logsumexp(pi_adj)
            LL_t = den - re_t
            
            # Normalize the updated state probabilities
            pi_t_next = (pi_adj + den).reshape([len(self.pM)] * self.k)
            
            # Carry forward updated state probabilities
            return (pi_t_next, LL + LL_t), pi_t_next
        
        # out of sample forecasting case
        if self.fitted:
            # for the first iteration
            if pi_start is None: pi_t = self.log_pi_t
            else: pi_t = pi_start
            r_eta = jnp.log(eta)
            x = self._x(self.beta0, self.beta1, self.eta, newd, oldd)
            states = self.states
            w_val = self.w(x, states, sigma_l)
            gamma_vec = self.gamma_vec
            A_tens = self.A_tens
            
        #otherwise the model is initialized
        else:
            # create the ergotic distribution to a tensor
            states, pi_t = self.ergotic_distr(support)
            # compute multifractal component
            x = self._x(beta0, beta1, eta)
            # compute level component in log-space
            r_eta = jnp.log(self.ts[1:]**eta)
            #compute observation likelihood matrix
            w_val = self.w(x, states, sigma_l)
            # Compute the transition tensor
            gamma_vec = self.gamma_prob(b, gamma_k)
            A_tens = self.A_tensor(gamma_vec)

        if (jnp.shape(w_val) == () or jnp.shape(w_val)[0] == 1):
            # Single observation: call transition_tensor once
            (pi_t, fin_LL), pi_list = transition_tensor((pi_t, 0.0), (w_val, r_eta))
            pi_list = jnp.expand_dims(pi_list, 0)  # to keep output shape consistent
        else:
            # Sequence: use scan
            (pi_t, fin_LL), pi_list = jax.lax.scan(transition_tensor, (pi_t, 0.0), (w_val, r_eta))
                     
        if comp: return pi_t, pi_list, fin_LL
        else: return fin_LL
    
    def fit(self, init_param, maxiter=500, verb=True, method="Nelder-Mead"):
        """
        init_param: 
            log_sigma_l,               # scalar
            log_b_minus1,              # scalar
            logit_gamma_K,             # scalar (for the largest state K)
            log_eta,                   # scalar
            beta0,                     # scalar, unconstrained
            beta1,                     # scalar, unconstrained
            *log_m_raw                 # length = len(self.pM)
        """

        M = len(self.pM)

        def unpack_params(x):
            #slice out params
            log_sigma_l   = x[0]
            log_b_minus1  = x[1]
            logit_gamma_K = x[2]
            log_eta       = x[3]
            beta0         = x[4]
            beta1         = x[5]
            log_m_raw     = x[6:]

            #back‐transform
            sigma_l = jnp.exp(log_sigma_l)
            b       = 1.0 + jnp.exp(log_b_minus1)
            gamma_k = jax.nn.sigmoid(logit_gamma_K)
            eta = jnp.exp(log_eta)
            m_pos = jnp.exp(log_m_raw)
            support = m_pos / jnp.dot(self.pM, m_pos)

            return gamma_k, b, sigma_l, beta0, beta1, eta, support

        def loss_fn(x):
            gamma_k, b, sigma_l, beta0, beta1, eta, support = unpack_params(x)
            # Defensive checks for invalid values
            if (
                not jnp.all(jnp.isfinite(support)) or
                not jnp.all(support > 0) or
                not jnp.isfinite(sigma_l) or sigma_l <= 0 or
                not jnp.isfinite(b) or b <= 1 or
                not jnp.isfinite(eta) or eta <= 0 or
                not jnp.isfinite(beta0) or
                not jnp.isfinite(beta1)
            ):
                return 1e10
            ll = self.update_distr(sigma_l, b, gamma_k, support, beta0, beta1, eta, comp=False)
            if not jnp.isfinite(ll):
                return 1e10
            return ll

        if method == "LBFGSB":
            bounds = (jnp.repeat(-100, M+3), jnp.repeat(100, M+3))
            solver = LBFGSB(fun=loss_fn, maxiter=maxiter, verbose=verb)
            result = solver.run(init_param, bounds=bounds)
            self.converged = result.state.error < solver.tol
            if not self.converged: print("Warning! Optimization did not converge.")
            
        elif method == "Nelder-Mead":
            solver = ScipyMinimize(method="Nelder-Mead",fun=loss_fn, maxiter=maxiter, options={"disp": verb})
            result = solver.run(init_param)
            self.converged = result.state.success
            if not self.converged: print("Warning! Optimization did not converge.")
            self.n_iter = result.state.iter_num
        
        elif method == "differential_evolution":
            from scipy.optimize import differential_evolution

            # Define bounds for each parameter (adjust as needed)
            bounds = [
                (-10, 5),   # log_sigma_l
                (-10, 5),   # log_b_minus1
                (-5, 5),    # logit_gamma_K
                (-5, 5),    # log_eta
                (-10, 10),  # beta0
                (-10, 10)   # beta1
            ]
            bounds += [(-5, 5)] * M  # log_m_raw for each support

            # DE expects numpy arrays, so wrap loss_fn
            def loss_fn_np(x):
                x = jnp.array(x)
                return float(loss_fn(x))

            result = differential_evolution(loss_fn_np, bounds, maxiter=maxiter, disp=verb)
            self.converged = result.success
            if not self.converged:
                print("Warning! Optimization did not converge.")
            # Mimic the result object for consistency
            class DummyResult:
                pass
            dummy = DummyResult()
            dummy.params = jnp.array(result.x)
            self.result = dummy
            result = dummy
        
        self.result = result
        
        self.gamma_k, self.b, self.sigma_l, self.beta0, self.beta1, self.eta, self.support = unpack_params(result.params)
        self.log_pi_t, self.pi_list, self.LL = self.update_distr(self.sigma_l, self.b, self.gamma_k, self.support, self.beta0, self.beta1, self.eta)
        
        # exit log world and flatten probabilitis
        self.pi_list = jnp.exp(jnp.stack([pi.flatten() for pi in self.pi_list]))
        
        self.pi_t = jnp.exp(self.log_pi_t)
        self.gamma_vec = self.gamma_prob(self.b, self.gamma_k)
        self.A_tens = self.A_tensor(self.gamma_vec)
        self.x = self._x(self.beta0, self.beta1, self.eta)
        self.states = jnp.array(list(itertools.product(self.support, repeat=self.k)))
        #Maximized posterior marginals corresponding states (different from Viterbi-implied states)
        self.MPM_states = self.states[jnp.argmax(self.pi_list, axis=1)]
        
        self.sigma = self.sigma_l * jnp.sqrt(jnp.prod(self.states, axis=1))
        self.fitted = True
    
    def forecast(self, step=1, insample=True, new_lpi=None, newd=None):
        """
        Returns expected volatiliy (standard deviation) after step periods of time.
        It is then possible to simulate returns by sampling from N(0, exp_std)
            :param step: How many time units into the future to forecast
            :param insample: Specifies whether fitted distribution needs to be used.
            :param new_lpi: if insample is false, specifies the new pi_t to use
        """
        
        if self.fit:
            
            if insample: 
            # initialize the predictive distribution as current distribution
                log_pi_forecast = self.log_pi_t
            #first observation is last of original data
                r_n = self.ts[-1]
            else: 
                log_pi_forecast = new_lpi
                r_n = newd
            
            #initialize vector of new data
            r_new = jnp.zeros(step)    
            
            for h in range(step):
                # Apply the transition tensor to get the forecast probabilities
                for k in range(self.A_tens.shape[0]):
                    # Move mode-k axis to front
                    log_A_k = self.A_tens[k].T
                    log_pi_forecast = jnp.moveaxis(log_pi_forecast, k, 0) 

                    # For each j, sum over i: logsumexp(log_A_k[i, j] + log_pi_t[i])
                    log_pi_forecast = jax.vmap(  # over j in log_A_k.T
                        lambda log_col: logsumexp(log_col[:, None] + log_pi_forecast, axis=0)
                    )(log_A_k)
                    
                    log_pi_forecast = jnp.moveaxis(log_pi_forecast, 0, k)
            
                # flatten and exp() to obtain the probability vector
                pi_forecast = jnp.exp(log_pi_forecast.flatten())
                x_new = jnp.sum(pi_forecast * self.sigma)
                r_n = self.beta0 + self.beta1 * r_n+ x_new*r_n**self.eta + r_n
                
                #apply change
                r_new.at[h].set(r_n)
            
            return r_n, x_new
        
        else: raise ValueError("Fit model first.")
        
    def bayes_filter(self, newd):
        """
        Makes 1-step ahead out of sample forecast and then updates distribution with new data.
        """
        new_lpi = self.log_pi_t
        x_list = []
        r_list = []
        oldd = self.ts[-1]
        
        for y in newd:
            # Update state with new observation first
            new_lpi, _, _ = self.update_distr(self.sigma_l, self.b, self.gamma_k, self.support, self.beta0, self.beta1, self.eta, newd=y, oldd=oldd, pi_start=new_lpi)
            # Forecast using the updated state
            r_t, x_new = self.forecast(insample=False, new_lpi=new_lpi, newd=y)
            r_list.append(r_t)
            x_list.append(x_new)
            oldd = r_t
        return r_list, x_list
            
    def tail_idx(self, nsim = 10):
        """
        Computes the tail index of the time series and of model simulations
        """
        if self.fitted:
            self.tail = pareto_shape(self.ts)
            
            #initialize list of simulation tails
            sim_tail = jnp.zeros(shape=[nsim, len(self.tail)])
            
            # Compute the pareto tails for each simulation and store in each row of sim_tail
            for i in range(nsim):
                r_sim, _ = sim_levelMSM(self.support, self.pM, self.sigma_l, self.b, self.gamma_k, self.k, self.beta0, self.beta1, self.eta, len(self.ts))
                sim_tail = sim_tail.at[i].set(pareto_shape(r_sim))
                
            # compute the average of each tail order
            sim_tail = jnp.mean(sim_tail, axis=0)
            self.tail_pv = tail_pv(self.tail, sim_tail)
            
        else: raise ValueError("Fit model first.")



import numpy as np

def partition_map(arr, func=None, num_partitions=2):
    """
    Given a function that maps an array to a real value, it computes an array containing the
    the output of that function for a specified number of ordered partition of an array
    
    :param arr: the input array to compute the partition-wise function
    :param func: a function that maps an array to a real value
    :param num_partitions: n of partitions to split the array in
    """
    
    if func is None:
        func = lambda x: 0    
    
    partition_size = len(arr) // num_partitions
    result = []
    
    for i in range(num_partitions):
        start = i * partition_size
        end = start + partition_size if i < num_partitions - 1 else len(arr)  # Ensure last partition includes remaining elements
        partition = arr[start:end]
        result.append(func(partition))
    
    return np.array(result)

def OLS(x, y, intercept=True):
    """
    Returns array of weights of an OLS regression 
    """
    x = (np.array(x))
    y = np.array(y)
    if(intercept): x = np.c_[x, np.ones(x.shape[0])]
    return np.dot(np.linalg.inv(np.dot(x.T, x)), np.dot(x.T, y))



""" Module for the data likelihood/emissions of the HMM. """

import jax.numpy as jnp
import jax.scipy.stats as jss

from abc import abstractmethod

class data_likelihood:

    def __init__(self, ts: jnp.ndarray) -> None:
        self.ts = ts
    
    @abstractmethod
    def likelihood(self, latent_states: jnp.ndarray)->jnp.ndarray:
        self.latent_states = latent_states

#TODO: generalize to work for any distribution in jss

class dlk_normal(data_likelihood):
    def likelihood(self, latent_states: jnp.ndarray)->jnp.ndarray:
        super().likelihood(latent_states)
        return jss.norm.pdf(self.ts.reshape(-1, 1), loc=0, scale=self.latent_states)


# Contains abstract time series classes. 
# Stochastic classes have an additional method containing simulation.


import numpy as np
import pandas as pd

import jax.numpy as jnp
from abc import abstractmethod, ABC

class time_series(ABC):
    """Generic time series class."""
    
    def __init__(self, ts: np.ndarray | jnp.ndarray | pd.Series | pd.DataFrame, name: str | None = None) -> None:
        """
        Initializes the model, data to use, hyperparameters, ...
        ts: time series to initialize the model to.
        """
        self.ts = jnp.array(ts)
        self.name = name or self.__class__.__name__
    
    @abstractmethod
    def fit(self) -> jnp.ndarray:
        """Returns the model's parameters using a loss function."""
        pass
    
    @abstractmethod
    def forecast(self, filtration) -> jnp.ndarray:
        """Makes a prediction with the fitted model.
        filtration: horizon of forecast or new independent variables to use
        """
        pass
            
class stochastic_ts(time_series):
    """Time series with random components"""

    @abstractmethod
    def simulation(self, nsim: int)-> jnp.ndarray:
        """Generates nsim datapoints from the model"""
        pass


""" Module for the initial distribution of the HMM. """

import jax.numpy as jnp
import itertools

from abc import abstractmethod

class initial_distr:
    """
    The initial distribution of the process. Either the ergotic distribution, 
    or another initial choice based on context-criteria.
        
    It can also be learned from data if it depends on parameters.
    """
    def __init__(self, marg_prob_mass:jnp.ndarray, num_latent:int= 1) -> None:
        
        if jnp.isclose(jnp.sum(marg_prob_mass), 1.0, atol=1e-6):
            self.num_latent = num_latent
            self.marg_prob_mass = marg_prob_mass
        else: raise ValueError("The marginal probability mass needs to sum to 1.")

# TODO: add generalized discrete distribution function with dependence function.

class homogeneous_independent(initial_distr):
    """
    Class for initial joint distribution of independent latent states with the same marginal distribution (homogeneous).
    """
    
    @abstractmethod
    def support(self, marg_support:jnp.ndarray) -> jnp.ndarray:
        """
        Joint support of the distribution.
        At this level, only the disjoint states are computed. The aggregation type is model dependent.
        """
        self.support_disjoint = jnp.array(list(itertools.product(marg_support, repeat=self.num_latent)))
    
    def mass(self) -> jnp.ndarray:
        """
        Joint initial probability mass. The joint probabilities are the row-wise product of the cartesian product 
        repeated as the number of latent states - the dependece term (implemented in concretization classes).
        
        Returns it in a tensor form, ready for tensor-forward algorithm.
        """
        prob_disjoint = jnp.array(list(itertools.product(self.marg_prob_mass, repeat=self.num_latent)))
        return jnp.prod(prob_disjoint, axis=1)
    
class multiplicative_cascade(homogeneous_independent):
    """
    Initial distribution given by the square root of the product of the marginal states and a positive variable.
    """
    def support(self, uncond_term:float, marg_support:jnp.ndarray)->jnp.ndarray:
        super().support(marg_support)
        return uncond_term * jnp.sqrt(jnp.prod(self.support_disjoint, axis=1))


from ..helper import OLS
from ..descriptive import acf
import numpy as np

class ar:
    
    def __init__(self, ts, p=0):
        
        # length and lag
        self.ts = np.array(ts)
        self.lag = p
        
        #constructuion of independent and dependent variables
        self.y = ts[p:]
        self.x = np.array([ts[i:i+p] for i in range(len(ts) - p)])
        self.x = np.hstack((self.x, np.ones((self.x.shape[0], 1))))
        self.w = OLS(x=self.x, y=self. y, intercept=False)
        self.pred = np.dot(self.x, self.w)
        self.resid = self.y - self.pred
        self.rsq = 1 - np.sum(self.resid**2)/np.sum((self.y-np.nanmean(self.y))**2)
        self.rsq_adj = 1 - np.sum(self.resid**2*(len(self.y)-1))/np.sum((self.y-np.nanmean(self.y))**2*(len(self.y)-self.lag-1))
        self.acf = acf(self.ts).iloc[:self.lag, :]
        
    def forecast(self, t):
        x_new = self.ts[len(self.ts)-self.lag:]
        y_new = np.empty(t)
        
        for i in range(t):
            x_interc = np.concatenate(([1], x_new))
            y_new[i] = np.dot(x_interc, self.w)
            x_new = np.roll(x_new, shift=-1)
            x_new[-1] = y_new[i]
        return y_new



import jax
import jax.numpy as jnp
from jaxopt import LBFGSB
from functools import partial
from jaxopt import ScipyMinimize
from fractrics.pending_refactor.unscent_KF import *
from scipy.optimize import differential_evolution

class CascadeDTSM:
    """
    Cascade dynamic term structure model as in Calvet, Fisher, and Wu (2018).
    """
    def __init__(self, rates, n, taus, dt):
        self.n = n
        #series of the observed OIS rates
        self.rates = rates
        #terms of the rates
        self.terms = taus
        #time lag
        self.dt = dt
        
        if len(self.terms) != self.rates.shape[1]:
            raise ValueError(
            f"Number of maturities (len(terms)={len(self.terms)}) "
            f"must equal number of observed rates (rates.shape[1]={self.rates.shape[1]})"
            )
    
    @partial(jax.jit, static_argnames=["self", "mat"])
    def _kappas(self, k1, b, mat=False):
        """
        Speed of factors.
        :param mat: if True computes kappas in the lower-triangular form, otherwise returns a vector
        """
        kappas = k1*b**(jnp.arange(start=1, stop=self.n+1)-1)
        if mat: return jnp.diag(kappas) + jnp.diag(-kappas[1:], k=-1)
        else: return kappas
    
    @partial(jax.jit, static_argnames=["self"])    
    def _theta_q(self, theta_r, gamma, sigma, kappas):
        """risk-neutral drift adjustments"""
        return theta_r-gamma*sigma**2/kappas
    
    @partial(jax.jit, static_argnames=["self"])    
    def _fx(self, x_prev, dt, params):
        """
        Latent state-space dynamics. Error matrix is dealt with inside the Unscent Kalman filter.
        :param params: State dynamic parameters in specific order, such that:
            element 0: k1,
            element 1: b,
            element 2:theta_r
        """
        k1, b, theta_r = params
        kappam = self._kappas(k1, b, mat=True)
        #use matrix exponential, not element wise exp
        Phi = jax.scipy.linalg.expm(-dt*kappam)
        A = (jnp.identity(self.n) - Phi) @ jnp.full(self.n, theta_r)
        x_next = A + Phi @ x_prev
        
        return x_next
    
    @partial(jax.jit, static_argnames=["self"])    
    def zcb_price(self, tau, k1, theta_r, sigma, kappas, gamma, x):
        
        n = self.n

        num = jnp.cumprod(kappas[::-1])[::-1]
        num_mat = num[:, None]
        idx = jnp.arange(n)
        J2 = idx[:, None, None]
        I2 = idx[None, :, None]
        M2 = idx[None, None, :]
        mask3 = (M2 >= J2) & (M2 != I2)
        D3 = kappas[M2] - kappas[I2]
        prod_diff = jnp.prod(jnp.where(mask3, D3, 1.0), axis=2)

        # final denominator matrix:
        den = (kappas[None, :] * kappas[:, None]) * prod_diff

        # alpha matrix:
        alpha_mat = num_mat / den
        mask_load = idx[None, :] >= idx[:, None]
        alpha_mat = alpha_mat * mask_load

        # price loadings b (integral of a)
        M = 1.0 - jnp.exp(-kappas * tau)
        ploads = alpha_mat @ M

        # Intercept c
        G = tau - (1 - jnp.exp(-kappas * tau)) / kappas

        l1 = theta_r * k1 * jnp.dot(alpha_mat[:, 0], G)
        l2 = gamma * sigma**2 * jnp.sum(alpha_mat.T @ G)

        kappas_sum = kappas[:, None] + kappas[None, :]
        H = (
            tau
            - G[:, None]
            - G[None, :]
            + (1 - jnp.exp(-kappas_sum * tau)) / kappas_sum
        )                                                      

        quad = jax.vmap(lambda a_j: a_j @ (H @ a_j))(alpha_mat)
        l3 = 0.5 * sigma**2 * jnp.sum(quad)

        intercept = l1 - l2 - l3
        
        return jnp.exp(-jnp.dot(ploads, x) - intercept)

    @partial(jax.jit, static_argnames=["self"])
    def _hois(self, x, params):
        """
        Observation function using OIS rates.
        :param params: touple of parameters:
        """
        b, k1, theta_r, gamma, sigma = params
        kappas = self._kappas(k1, b)
        zcb_price_vmap = jax.vmap(lambda tau: self.zcb_price(tau, k1, theta_r, sigma, kappas, gamma, x))

        bond = zcb_price_vmap(self.terms)

        OIS = -(1/self.terms)*jnp.log(bond)
        return OIS

    def fit(self, init_param, maxiter=500, verb=True, method="Nelder-Mead"):

        def unpack_params(x):
            #slice out params
            log_k1   = x[0]
            log_sigma = x[1]
            log_b_minus1  = x[2]
            theta_r = x[3]  # no constraint
            gamma = x[4]    # no constraint
            log_sigma_e = x[5]

            #back‐transform
            k1 = jnp.exp(log_k1)
            sigma = jnp.exp(log_sigma)
            sigma_e = jnp.exp(log_sigma_e)
            b = 1.0 + jnp.exp(log_b_minus1)

            return k1, sigma, b, theta_r, gamma, sigma_e

        def loss_fn(params):
            k1, sigma, b, theta_r, gamma, sigma_e = unpack_params(params)
            fx_params = (k1, b, theta_r)
            hx_params = (b, k1, theta_r, gamma, sigma)

            # observation error covariance
            R = jnp.eye(self.rates.shape[1])*(sigma_e**2)
            # X noise covariance
            Q = jnp.eye(self.n)

            # unscent Kalman filter set-up
            alpha, beta, kappa = 1.0, 2.0, 0.0
            lam = alpha**2 * (self.n + kappa) - self.n
            c   = self.n + lam
            Wm0 = lam / c
            Wc0 = Wm0 + (1 - alpha**2 + beta)
            W_i = 1.0 / (2.0 * c)
            Wm = jnp.full((2*self.n + 1,), W_i).at[0].set(Wm0)
            Wc = jnp.full((2*self.n + 1,), W_i).at[0].set(Wc0)

            init_state = UKFState(x=jnp.zeros(self.n), P=jnp.eye(self.n))

            ukf_args = (self._fx, fx_params, self._hois, hx_params,
                Q, R, Wm, Wc, self.dt, c)
            
            def scan_fn(st, y):
                new_st, negll = ukf_step(st, y, *ukf_args)
                return new_st, (new_st, negll)
            
            #computes likelihood for given rates
            _, (states, nlls) = jax.lax.scan(
                scan_fn,
                init_state,
                self.rates
            )
            
            return jnp.sum(nlls)

        #bounded gradient optimization
        if method=="LBFGSB":

            bounds = (jnp.repeat(-1000, init_param.shape[0]), jnp.repeat(1000, init_param.shape[0]))
            solver = LBFGSB(fun=loss_fn, maxiter=maxiter, verbose=verb, implicit_diff=False, history_size=5)
            result = solver.run(init_param, bounds=bounds)
            self.result = result
            self.converged = result.state.error < solver.tol
            self.fitted = True

        elif method=="Nelder-Mead":
            solver = ScipyMinimize(method="Nelder-Mead",fun=loss_fn,maxiter=maxiter, options={"disp": verb})
            self.result = solver.run(init_param)
            self.num_iter = self.result.state.iter_num
            self.converged = self.result.state.success
            self.ll = self.result.state.fun_val
        
        elif method == "differential_evolution":

            # Define bounds for each parameter (adjust as needed)
            bounds = [(-5, 5),
                      (-5, 5), 
                      (-5, 5),
                      (-5, 5),
                      (-5, 5)] 

            # DE expects numpy arrays, so wrap loss_fn
            def loss_fn_np(x):
                x = jnp.array(x)
                return float(loss_fn(x))

            result = differential_evolution(loss_fn_np, bounds, maxiter=maxiter, disp=verb)
            self.converged = result.success
            if not self.converged:
                print("Warning! Optimization did not converge.")
            # Mimic the result object for consistency
            class DummyResult:
                pass
            dummy = DummyResult()
            dummy.params = jnp.array(result.x)
            self.result = dummy
            result = dummy
            
        else: raise ValueError("Selected method not supported.")
        
        if not self.converged: print("Warning! Optimization did not converge.")
        self.fitted = True
        self.k1, self.sigma, self.b, self.theta_r, self.gamma, self.sigma_e = unpack_params(self.result.params)

    def manual_fit(self, params, filtered_xT, ll, converged) :
        self.k1, self.sigma, self.b, self.theta_r, self.gamma, self.sigma_e = params
        self.ll = ll
        self.fitted = True
        self.converged = converged

    def filter(self):
        """
        Runs the Unscented Kalman Filter over the observed rate series
        and returns the full sequence of filtered UKFState objects.
        """

        if not hasattr(self, 'fitted') or not self.fitted:
            raise RuntimeError("Model must be fitted before filtering.")

        Q = jnp.eye(self.n)
        R = jnp.eye(self.rates.shape[1]) * self.sigma_e**2

        alpha, beta, kappa = 1.0, 2.0, 0.0
        lam = alpha**2 * (self.n + kappa) - self.n
        c = self.n + lam
        Wm0 = lam / c
        Wc0 = Wm0 + (1 - alpha**2 + beta)
        Wi = 1 / (2 * c)
        Wm = jnp.full((2 * self.n + 1,), Wi).at[0].set(Wm0)
        Wc = jnp.full((2 * self.n + 1,), Wi).at[0].set(Wc0)
        init_state = UKFState(x=jnp.zeros(self.n), P=jnp.eye(self.n))
        fx_args = (self._fx, (self.k1, self.b, self.theta_r))
        hx_args = (self._hois, (self.b, self.k1, self.theta_r, self.gamma, self.sigma))

        def filter_step(state, y):
            new_state, _ = ukf_step(
                state, y,
                fx_args[0], fx_args[1],
                hx_args[0], hx_args[1],
                Q, R, Wm, Wc, self.dt, c
            )
            return new_state, new_state

        _, filtered_states = jax.lax.scan(
            filter_step,
            init_state,
            self.rates
        )

        return filtered_states

    def forecast_zcb(self,
                     x0: jnp.ndarray,
                     steps: int,
                    ) -> jnp.ndarray:
        """
        Forecast zero-coupon bond prices for `self.terms` maturities, over `steps`
        future time‐points, starting from latent state `x0`.
        Returns an array of shape (steps, n_maturities).
        """
        
        # fx parameters under P
        fx_params = (self.k1, self.b, self.theta_r)

        def one_step(x_prev, _):
            #Predict next latent state (conditional mean)
            x_next = self._fx(x_prev, self.dt, fx_params)

            #Compute ZCB prices at x_next for all self.terms
            kappas = self._kappas(self.k1, self.b)
            zcb_one = jax.vmap(
                lambda tau: self.zcb_price(
                    tau,
                    self.k1,
                    self.theta_r,
                    self.sigma,
                    kappas,
                    self.gamma,
                    x_next
                )
            )(self.terms)

            return x_next, zcb_one

        # scan over `steps` iterations
        _, zcb_paths = jax.lax.scan(
            one_step,
            x0,
            None,
            length=steps
        )
        return zcb_paths
        



from fractrics._ts_components._HMM._base import HMM

import fractrics._ts_components._HMM._forward as _forward
import fractrics._ts_components._HMM._transition_tensor as _transition_tensor
import fractrics._ts_components._HMM._data_likelihood as _data_likelihood
import fractrics._ts_components._HMM._initial_distribution as _initial_distribution

from functools import partial

from jax.lax import scan
from jax.nn import softplus, sigmoid
from jax import jit #,hessian, jacfwd, vmap # for standard errors

import numpy as np

import jax.numpy as jnp
import jax.random as random

from jaxopt import BFGS

# TODO: 
# create a custom class to handle parameters and data easier in production. Have it make a summary
# add standard errors of parameters

class MSM(HMM):
    """Univariate Discrete Markov Switching Multifractal model."""
    
    def __init__(self, ts: np.ndarray | jnp.ndarray,
                 num_latent: int = 1,
                 marg_prob_mass = jnp.full(2, 0.5),
                 name: str | None = None,
        ) -> None:
        
        if (jnp.any(ts <= 0 | jnp.isinf(ts))): raise ValueError("MSM is defined for positive, finite time series only.")
        else:
            self.r = jnp.log(ts[1:])-jnp.log(ts[:-1])
            self.marg_prob_mass = marg_prob_mass #necessary to correctly create optimization constrains

            super().__init__(ts=ts,
                num_latent=num_latent,
                initial_dist= _initial_distribution.multiplicative_cascade(num_latent=num_latent, marg_prob_mass=marg_prob_mass),
                transition_tensor = _transition_tensor.poisson_arrival(num_latent=num_latent, marg_prob_mass=marg_prob_mass),
                data_likelihood=_data_likelihood.dlk_normal(ts=self.r),
                forward=_forward.factor_transition(num_latent=num_latent),
                name=name)
    
    def fit(self, initial_parameters, maxiter:int, verbose=False):
        """
        
        Parameters to be optimized, stored in initial_parameters in the following order:
            uncond_term: unconditional variance component of the latent states: > 0
            arrival_gdistance: geometric distrance between each Poisson arrival: > 0
            hf_arrival: the high-frequency Poisson arrival: between 0 and 1
            marg_support: (rest of the parameters vector) the support of the marginal distribution: >0 and unity expectation.
        """
        
        ergotic_dist  = self._initial_dist.mass()
        
        def reparameterization(params):
            """Enforces constrains on parameters before input to the solver."""

            positive_constraint = softplus(params[:2])
            possion_constraint = sigmoid(params[2])
            
            support = params[3:]

            support_positive = jnp.exp(support)
            support_constraint = support_positive / jnp.dot(self.marg_prob_mass, support_positive)
            
            return jnp.concatenate([positive_constraint, jnp.array([possion_constraint]), support_constraint]) # type: ignore
        
        def loss_fn(params):
            
            constrained_params = reparameterization(params)
            
            uncond_term=constrained_params[0]
            arrival_gdistance=constrained_params[1]
            hf_arrival=constrained_params[2]
            marg_support = constrained_params[3:]

            latent_states = self._initial_dist.support(uncond_term=uncond_term, marg_support=marg_support) # type: ignore
            data_likelihood = self._data_likelihood.likelihood(latent_states=latent_states)
            transition_tensor = self._transition_tensor.t_tensor(arrival_gdistance=arrival_gdistance, hf_arrival=hf_arrival) # type: ignore
            
            NLL, distr_fin, distr_list = self._forward.update(ergotic_dist, data_likelihood, transition_tensor) # type: ignore
            
            return NLL, (distr_fin, transition_tensor, distr_list)

        solver = BFGS(fun=loss_fn, has_aux=True, maxiter=maxiter, verbose=verbose)
        result = solver.run(init_params=initial_parameters)
        
        params_optimized = reparameterization(result.params)
        current_distribution = result.state.aux[0]
        transition_tensor = result.state.aux[1]
        distribution_history = result.state.aux[2]
        negative_log_likelihood = result.state.value
        
        #TODO: add info in custom class for summary
        
            #gradient = result.state.grad
            #number_iteration = result.state.iter_num
            #MPM
            #viterbi
            #other useful information
            
            #standard error (not tested):
                # nll_hessian = hessian(negative_log_likelihood)(params_optimized, self.r)
                # covariance_matrix = jnp.linalg.inv(nll_hessian)
                # standard_errors = jnp.sqrt(jnp.diag(covariance_matrix))
            #robust standard errors (not tested):
                # assume nll_observation is nll not summed (ony need to modify scan above)
                # observation_jacobian = jacfwd(nll_observation)
                # score_matrix = vmap(observation_jacobian, in_axes=(None, 0))(params_optimized, self.r)
                # outer_product_scores = jnp.dot(score_matrix.T, score_matrix)
                # robust_cov = jnp.dot(jnp.dot(covariance_matrix, outer_product_scores),covariance_matrix)
                # robust_se = jnp.sqrt(jnp.diag(robust_cov))
                                
        return params_optimized, current_distribution, transition_tensor, distribution_history, negative_log_likelihood
    
    #TODO: take custom class as input to not decompose parameters
    @partial(jit, static_argnames=["self", "number_simulations"]) #number_sim needs to be static for reshape of keys
    def simulation(self, 
            number_simulations:int,
            poisson_arrivals,
            marginal_support,
            unconditional_term,
            key = random.PRNGKey(0))->tuple[jnp.ndarray, jnp.ndarray]:

        key, key_init = random.split(key)

        # Draw the initial states from marginal distribution
        init_states = random.choice(key_init, marginal_support, (self.num_latent,), p=self.marg_prob_mass)

        # prepare nsim sets of keys for simulations
        keys = random.split(key, number_simulations * 3).reshape(number_simulations, 3, 2)
        
        def _step(states, key_triple):

            key_arrival, key_switch, key_noise = key_triple
            switch_mask = random.bernoulli(key_arrival, p=poisson_arrivals)
            new_vals = random.choice(key_switch, marginal_support, (self.num_latent,), p=self.marg_prob_mass)
            
            states = jnp.where(switch_mask, new_vals, states)
            vol = unconditional_term * jnp.sqrt(jnp.prod(states)) # type: ignore
            r = vol*random.normal(key_noise)
            
            return states, (vol, r)
        
        _, (volatility_sim, return_sim) = scan(_step,init_states, keys)
        
        return return_sim, volatility_sim
    
    #TODO: generalize in base HMM class
    def forecast(self, horizon:int, prior:jnp.ndarray, *transition_matrices: jnp.ndarray):
        return self._forward.forecast(horizon, prior, *transition_matrices) #type: ignore


"""Statistics to test performances of models"""

import jax.numpy as jnp
import jax.scipy.stats as jss

def pareto_shape(ts: jnp.ndarray, order: jnp.ndarray | None) -> jnp.ndarray:
    """Computes the Pareto Shape index as the inverse of Hill's tail index of a time series."""
    ts = jnp.asarray(ts)
    ordst = jnp.sort(ts)
    n = ordst.shape[0]

    if order is None:
        order = jnp.array([n * p // 100 for p in (1, 5, 7, 10)], dtype=jnp.int32)
    order = jnp.atleast_1d(order).astype(jnp.int32)

    idx = jnp.arange(n)
    mask = idx[None, :] >= order[:, None]

    ref = ordst[order]
    ref = ref[:, None]

    tail = jnp.where(mask, ordst[None, :], 1.0)
    logs = jnp.log(tail / ref)

    hill = jnp.sum(logs, axis=1) / order
    return 1.0 / hill

def tail_pv(ps1, ps2, order=[1, 5, 7 , 10]):
    """
    Computes the p-value/array of p-values of 2 pareto-shape statistics
    """
    order = jnp.array(order)
    if ps1.shape != ps2.shape or ps1.shape != order.shape: raise ValueError("order and Pareto statistics must be of same length")
    
    nabsZ = -jnp.abs((ps1 - ps2)*jnp.sqrt(order/2))
    return 2 * jss.norm.cdf(nabsZ, loc=0, scale=1)



from abc import abstractmethod
#from typing import Callable

import jax.numpy as jnp
from jax.lax import scan
from jax.nn import softmax
from jax.scipy.special import logsumexp

# TODO: add touple check using from typing import Tuple
# consider precision tuning (float32/bfloat16) for speedup
# build optimized engine for HMMs where emission factorize
# consider using predictive update directly in log space

class forward_update:
    """
    Abstract forward algorithm for discrete time HMMs. Inputs are assumed to not already be in log form.
    Distinct from continuous latent models, in which probability distributions cannot be represented using tensors.
    """
    def __init__(self,
                 num_latent: int = 1) -> None:
        """
        num_latent: number of latent variables
        num_states: number of states that each latent (homogeneous) variable can assume
        """

        self.num_latent = num_latent
    
    @abstractmethod
    def make_predictive_function(self):
        """Function that defines and returns the function for computing the predictive tensor."""
        pass
    
    @abstractmethod
    def update(self,
               distr_initial: jnp.ndarray,
               data_likelihood: jnp.ndarray,
               trans_tensor: jnp.ndarray
               ):
        """Function that updates the latent distribution along a matrix of emissions. Should use the function of make_predictive_function"""
        pass

class factor_transition(forward_update):
    """
    Forward algorithm for discrete time HMM with factorable latent states and joint emission.
    trans_tensor is supposed to be such that each of its dimension is the marginal transition matrix of each latent state.
    """
    
    @staticmethod # necessary to be used in update
    def make_predictive_function(*transition_matrices):
        """Returns a function that computes the predictive distribution in tensor form."""
        
        #NOTE: currently the forward algorithm breaks if given a jax tensor as input, due to the way enumerate handles the predictive update
        def make_predictive_distribution(prior:jnp.ndarray)->jnp.ndarray:
            
            dims = tuple(A.shape[0] for A in transition_matrices)
            predictive_tensor = prior.reshape(*dims)
            
            for axis, A in enumerate(transition_matrices):
                predictive_tensor = jnp.moveaxis(predictive_tensor, axis, -1)
                predictive_tensor = jnp.tensordot(predictive_tensor, A, axes=([-1], [0]))
                predictive_tensor = jnp.moveaxis(predictive_tensor, -1, axis)
            return predictive_tensor

        return make_predictive_distribution
    
    def update(self,
               distr_initial: jnp.ndarray,
               data_likelihood: jnp.ndarray,
               transition_matrices: tuple[jnp.ndarray]
              )-> tuple:
        
        predictive_function = self.make_predictive_function(*transition_matrices)
        
        small_constant = 1e-45
        
        dims = tuple(A.shape[0] for A in transition_matrices)
        log_tensor_initial_distribution = jnp.log(distr_initial.reshape(*dims) + small_constant)
        
        log_data_likelihood_tensor = jnp.log(data_likelihood.reshape((data_likelihood.shape[0],) + dims) + small_constant)
        
        #NOTE: predictive_tensor needs to run in non-log space
        #TODO: manage the forward with less swithches between log and non log space
        def step(carry, log_data_likelihood_row):
            log_prior, nl_loss_likelihood = carry
            
            log_predictive_tensor = jnp.log(predictive_function(softmax(log_prior)) + small_constant)
            
            log_nonnormalized_posterior = log_predictive_tensor + log_data_likelihood_row
            log_loss_likelihood = -logsumexp(log_nonnormalized_posterior)
            log_normalized_posterior = log_nonnormalized_posterior + log_loss_likelihood
            return (log_normalized_posterior, nl_loss_likelihood + log_loss_likelihood), log_normalized_posterior
        
        carry_initial = (log_tensor_initial_distribution, 0.0)
        
        (log_final_posterior, final_loss), log_distribution_list = scan(step, carry_initial, log_data_likelihood_tensor)
        return final_loss, jnp.exp(log_final_posterior), jnp.exp(log_distribution_list)
    
    #TODO: move in HMM base class by handling the specific inputs
    def forecast(self, horizon:int, prior: jnp.ndarray, *transition_matrices: jnp.ndarray):
        predictive_function = self.make_predictive_function(*transition_matrices)
        
        def step(carry, _):
            carry = predictive_function(carry)
            return carry, carry
        _, predictive_list = scan(step, prior, xs=None, length=horizon)
        
        return predictive_list


from typing import NamedTuple
import jax, jax.numpy as jnp
from functools import partial

class UKFState(NamedTuple):
    x: jnp.ndarray  # (dim_x,)
    P: jnp.ndarray  # (dim_x, dim_x)

def sigma_points(x, P, c):
    A = jnp.linalg.cholesky(c * P)
    dev = jnp.concatenate([A, -A], axis=1)   
    return jnp.vstack([x[None,:], (x + dev.T)])

def predict(x, P, fx, fx_args, Q, Wm, Wc, dt, c):
        
    sigmas = sigma_points(x, P, c)
       
    sigmas_f = jax.vmap(lambda s: fx(s, dt, fx_args))(sigmas)
    x_pred = jnp.dot(sigmas_f.T, Wm)
    diff = sigmas_f - x_pred  
    cross = jnp.einsum('bi,bj->bij', diff, diff)
    P_pred = Q + jnp.sum(Wc[:, None, None] * cross, axis=0)

    return x_pred, P_pred

def update(x, P, z, hx, hx_args, R, Wm, Wc, c):
    
    sigmas = sigma_points(x, P, c)
    sigmas_h = jax.vmap(lambda s: hx(s, hx_args))(sigmas)
    z_pred = jnp.dot(Wm, sigmas_h) 
    dz = sigmas_h - z_pred
    
    Pz = R + jnp.einsum('i,ij,ik->jk', Wc, dz, dz)

    # spectral projection for numerical stability
    Pz = 0.5 * (Pz + Pz.T)
    eigvals, eigvecs = jnp.linalg.eigh(Pz)
    min_eig = jnp.min(eigvals)
    eigvals_clipped = jnp.clip(eigvals, a_min=1e-6)
    Pz = (eigvecs * eigvals_clipped) @ eigvecs.T
    
    eps = 1e-6
    Pz = jnp.where(
        min_eig < eps,
        Pz + (eps - min_eig + 1e-8) * jnp.eye(Pz.shape[0]),
        Pz
    )

    dx = sigmas - x
    
    Pxz = jnp.einsum('i,ij,ik->jk', Wc, dx, dz)
    
    K = jnp.linalg.solve(Pz, Pxz.T).T
    
    x_new = x + K @ (z - z_pred)
    
    P_new = P - K @ Pz @ K.T
    P_new = 0.5*(P_new + P_new.T)
    eigvals, eigvecs = jnp.linalg.eigh(P_new)
    eigvals_clipped = jnp.clip(eigvals, a_min=1e-6)
    P_new = eigvecs @ jnp.diag(eigvals_clipped) @ eigvecs.T
    
    err = z - z_pred
    ll = -0.5*(jnp.linalg.slogdet(Pz)[1] + err.T @ jnp.linalg.inv(Pz) @ err + z.size*jnp.log(2*jnp.pi))
    
    return x_new, P_new, ll

@partial(jax.jit, static_argnames=('fx','hx'))
def ukf_step(state: UKFState, z, fx, fx_args, hx, hx_args, Q, R, Wm, Wc, dt, c):
    (x, P) = state

    x_pred, P_pred = predict(x, P, fx, fx_args, Q, Wm, Wc, dt, c)
    x_new, P_new, ll = update(x_pred, P_pred, z, hx, hx_args, R, Wm, Wc, c)
    return UKFState(x_new, P_new), -ll
